/**
 * Copyright (c) 2014, Facebook, Inc. All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */
'use strict';

require('mock-modules').autoMockOff();

describe('FakeTimers', function() {
  var FakeTimers;

  beforeEach(function() {
    FakeTimers = require('../FakeTimers');
  });

  describe('construction', function() {
    /* jshint nonew:false */
    it('installs setTimeout mock', function() {
      var global = {};
      new FakeTimers(global);
      expect(global.setTimeout).not.toBe(undefined);
    });

    it('installs clearTimeout mock', function() {
      var global = {};
      new FakeTimers(global);
      expect(global.clearTimeout).not.toBe(undefined);
    });

    it('installs setInterval mock', function() {
      var global = {};
      new FakeTimers(global);
      expect(global.setInterval).not.toBe(undefined);
    });

    it('installs clearInterval mock', function() {
      var global = {};
      new FakeTimers(global);
      expect(global.clearInterval).not.toBe(undefined);
    });

    it('mocks process.nextTick if on exists on global', function() {
      var origNextTick = function() {};
      var global = {
        process: {
          nextTick: origNextTick
        }
      };
      new FakeTimers(global);
      expect(global.process.nextTick).not.toBe(origNextTick);
    });

    it('doesn\'t mock process.nextTick if real impl isnt present', function() {
      var global = {};
      new FakeTimers(global);
      expect(global.process).toBe(undefined);
    });
  });

  describe('runAllTicks', function() {
    it('runs all ticks, in order', function() {
      var global = {
        process: {
          nextTick: function() {}
        }
      };

      var fakeTimers = new FakeTimers(global);

      var runOrder = [];
      var mock1 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock1');
      });
      var mock2 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock2');
      });

      global.process.nextTick(mock1);
      global.process.nextTick(mock2);

      expect(mock1.mock.calls.length).toBe(0);
      expect(mock2.mock.calls.length).toBe(0);

      fakeTimers.runAllTicks();

      expect(mock1.mock.calls.length).toBe(1);
      expect(mock2.mock.calls.length).toBe(1);
      expect(runOrder).toEqual(['mock1', 'mock2']);
    });

    it('does nothing when no ticks have been scheduled', function() {
      var nextTick = jest.genMockFn();
      var global = {
        process: {
          nextTick: nextTick
        }
      };

      var fakeTimers = new FakeTimers(global);
      fakeTimers.runAllTicks();

      expect(nextTick.mock.calls.length).toBe(0);
    });

    it('only runs a scheduled callback once', function() {
      var global = {
        process: {
          nextTick: function() {}
        }
      };

      var fakeTimers = new FakeTimers(global);

      var mock1 = jest.genMockFn();
      global.process.nextTick(mock1);
      expect(mock1.mock.calls.length).toBe(0);

      fakeTimers.runAllTicks();
      expect(mock1.mock.calls.length).toBe(1);

      fakeTimers.runAllTicks();
      expect(mock1.mock.calls.length).toBe(1);
    });

    it('cancels a callback even from native nextTick', function() {
      var nativeNextTick = jest.genMockFn();

      var global = {
        process: {
          nextTick: nativeNextTick
        }
      };

      var fakeTimers = new FakeTimers(global);

      var mock1 = jest.genMockFn();
      global.process.nextTick(mock1);
      fakeTimers.runAllTicks();
      expect(mock1.mock.calls.length).toBe(1);
      expect(nativeNextTick.mock.calls.length).toBe(1);

      // Now imagine we fast forward to the next real tick. We need to be sure
      // that native nextTick doesn't try to run the callback again
      nativeNextTick.mock.calls[0][0]();
      expect(mock1.mock.calls.length).toBe(1);
    });

    it('doesnt run a tick callback if native nextTick already did', function() {
      var nativeNextTick = jest.genMockFn();

      var global = {
        process: {
          nextTick: nativeNextTick
        }
      };

      var fakeTimers = new FakeTimers(global);

      var mock1 = jest.genMockFn();
      global.process.nextTick(mock1);

      // Emulate native nextTick running...
      nativeNextTick.mock.calls[0][0]();
      expect(mock1.mock.calls.length).toBe(1);

      // Ensure runAllTicks() doesn't run the callback again
      fakeTimers.runAllTicks();
      expect(mock1.mock.calls.length).toBe(1);
    });

    it('throws before allowing infinite recursion', function() {
      var global = {
        process: {
          nextTick: function() {}
        }
      };

      var fakeTimers = new FakeTimers(global);

      global.process.nextTick(function infinitelyRecursingCallback() {
        global.process.nextTick(infinitelyRecursingCallback);
      });

      expect(function() {
        fakeTimers.runAllTicks();
      }).toThrow(
        'Ran 100000 ticks, and there are still more! Assuming we\'ve hit an ' +
        'infinite recursion and bailing out...'
      );
    });
  });

  describe('runAllTimers', function() {
    it('runs all timers in order', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      var runOrder = [];
      var mock1 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock1');
      });
      var mock2 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock2');
      });
      var mock3 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock3');
      });
      var mock4 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock4');
      });

      global.setTimeout(mock1, 100);
      global.setTimeout(mock2, 0);
      global.setTimeout(mock3, 0);
      var intervalHandler = global.setInterval(function() {
        mock4();
        global.clearInterval(intervalHandler);
      }, 200);

      fakeTimers.runAllTimers();
      expect(runOrder).toEqual(['mock2', 'mock3', 'mock1', 'mock4']);
    });

    it('does nothing when no timers have been scheduled', function() {
      var nativeSetTimeout = jest.genMockFn();
      var global = {
        setTimeout: nativeSetTimeout
      };

      var fakeTimers = new FakeTimers(global);
      fakeTimers.runAllTimers();
    });

    it('only runs a setTimeout callback once (ever)', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      var fn = jest.genMockFn();
      global.setTimeout(fn, 0);
      expect(fn.mock.calls.length).toBe(0);

      fakeTimers.runAllTimers();
      expect(fn.mock.calls.length).toBe(1);

      fakeTimers.runAllTimers();
      expect(fn.mock.calls.length).toBe(1);
    });

    it('doesnt pass the callback to native setTimeout', function() {
      var nativeSetTimeout = jest.genMockFn();

      var global = {
        setTimeout: nativeSetTimeout
      };

      var fakeTimers = new FakeTimers(global);

      var mock1 = jest.genMockFn();
      global.setTimeout(mock1, 0);

      fakeTimers.runAllTimers();
      expect(mock1.mock.calls.length).toBe(1);
      expect(nativeSetTimeout.mock.calls.length).toBe(0);
    });

    it('throws before allowing infinite recursion', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      global.setTimeout(function infinitelyRecursingCallback() {
        global.setTimeout(infinitelyRecursingCallback, 0);
      }, 0);

      expect(function() {
        fakeTimers.runAllTimers();
      }).toThrow(
        'Ran 100000 timers, and there are still more! Assuming we\'ve hit an ' +
        'infinite recursion and bailing out...'
      );
    });
  });

  describe('runTimersToTime', function() {
    it('runs timers in order', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      var runOrder = [];
      var mock1 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock1');
      });
      var mock2 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock2');
      });
      var mock3 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock3');
      });
      var mock4 = jest.genMockFn().mockImpl(function() {
        runOrder.push('mock4');
      });

      global.setTimeout(mock1, 100);
      global.setTimeout(mock2, 0);
      global.setTimeout(mock3, 0);
      global.setInterval(function() {
        mock4();
      }, 200);

      // Move forward to t=50
      fakeTimers.runTimersToTime(50);
      expect(runOrder).toEqual(['mock2', 'mock3']);

      // Move forward to t=60
      fakeTimers.runTimersToTime(10);
      expect(runOrder).toEqual(['mock2', 'mock3']);

      // Move forward to t=100
      fakeTimers.runTimersToTime(40);
      expect(runOrder).toEqual(['mock2', 'mock3', 'mock1']);

      // Move forward to t=200
      fakeTimers.runTimersToTime(100);
      expect(runOrder).toEqual(['mock2', 'mock3', 'mock1', 'mock4']);

      // Move forward to t=400
      fakeTimers.runTimersToTime(200);
      expect(runOrder).toEqual(['mock2', 'mock3', 'mock1', 'mock4', 'mock4']);
    });

    it('does nothing when no timers have been scheduled', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runTimersToTime(100);
    });

    it('throws before allowing infinite recursion', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      global.setTimeout(function infinitelyRecursingCallback() {
        global.setTimeout(infinitelyRecursingCallback, 0);
      }, 0);

      expect(function() {
        fakeTimers.runTimersToTime(50);
      }).toThrow(
        'Ran 100000 timers, and there are still more! Assuming we\'ve hit an ' +
        'infinite recursion and bailing out...'
      );
    });
  });

  describe('reset', function() {
    it('resets all pending setTimeouts', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      var mock1 = jest.genMockFn();
      global.setTimeout(mock1, 100);

      fakeTimers.reset();
      fakeTimers.runAllTimers();
      expect(mock1.mock.calls.length).toBe(0);
    });

    it('resets all pending setIntervals', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      var mock1 = jest.genMockFn();
      global.setInterval(mock1, 200);

      fakeTimers.reset();
      fakeTimers.runAllTimers();
      expect(mock1.mock.calls.length).toBe(0);
    });

    it('resets all pending ticks callbacks', function() {
      var global = {
        process: {
          nextTick: function() {}
        }
      };
      var fakeTimers = new FakeTimers(global);

      var mock1 = jest.genMockFn();
      global.process.nextTick(mock1);

      fakeTimers.reset();
      fakeTimers.runAllTicks();
      expect(mock1.mock.calls.length).toBe(0);
    });

    it('resets current runTimersToTime time cursor', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      var mock1 = jest.genMockFn();
      global.setTimeout(mock1, 100);
      fakeTimers.runTimersToTime(50);

      fakeTimers.reset();
      global.setTimeout(mock1, 100);

      fakeTimers.runTimersToTime(50);
      expect(mock1.mock.calls.length).toBe(0);
    });
  });

  describe('runOnlyPendingTimers', function() {
    it('runs all timers in order', function() {
      var global = {};
      var fakeTimers = new FakeTimers(global);

      var runOrder = [];

      global.setTimeout(function cb() {
        runOrder.push('mock1');
        global.setTimeout(cb, 100);
      }, 100);

      global.setTimeout(function cb() {
        runOrder.push('mock2');
        global.setTimeout(cb, 0);
      }, 0);

      global.setInterval(function() {
        runOrder.push('mock3');
      }, 200);

      fakeTimers.runOnlyPendingTimers();
      expect(runOrder).toEqual([
        'mock2',
        'mock1',
        'mock3'
      ]);

      fakeTimers.runOnlyPendingTimers();
      expect(runOrder).toEqual([
        'mock2',
        'mock1',
        'mock3',

        'mock2',
        'mock1',
        'mock3'
      ]);
    });
  });

  describe('runWithRealTimers', function() {
    it('executes callback with native setTimeout', function() {
      var nativeSetTimeout = jest.genMockFn();
      var global = {setTimeout: nativeSetTimeout};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runWithRealTimers(function() {
        global.setTimeout();
      });
      expect(nativeSetTimeout.mock.calls.length).toBe(1);
      expect(global.setTimeout.mock.calls.length).toBe(0);
    });

    it('executes callback with native setInterval', function() {
      var nativeSetInterval = jest.genMockFn();
      var global = {setInterval: nativeSetInterval};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runWithRealTimers(function() {
        global.setInterval();
      });
      expect(nativeSetInterval.mock.calls.length).toBe(1);
      expect(global.setInterval.mock.calls.length).toBe(0);
    });

    it('executes callback with native clearTimeout', function() {
      var nativeClearTimeout = jest.genMockFn();
      var global = {clearTimeout: nativeClearTimeout};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runWithRealTimers(function() {
        global.clearTimeout();
      });
      expect(nativeClearTimeout.mock.calls.length).toBe(1);
      expect(global.clearTimeout.mock.calls.length).toBe(0);
    });

    it('executes callback with native clearInterval', function() {
      var nativeClearInterval = jest.genMockFn();
      var global = {clearInterval: nativeClearInterval};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runWithRealTimers(function() {
        global.clearInterval();
      });
      expect(nativeClearInterval.mock.calls.length).toBe(1);
      expect(global.clearInterval.mock.calls.length).toBe(0);
    });

    it('resets mock setTimeout after executing callback', function() {
      var nativeSetTimeout = jest.genMockFn();
      var global = {setTimeout: nativeSetTimeout};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runWithRealTimers(function() {
        global.setTimeout();
      });
      expect(nativeSetTimeout.mock.calls.length).toBe(1);
      expect(global.setTimeout.mock.calls.length).toBe(0);

      global.setTimeout();
      expect(nativeSetTimeout.mock.calls.length).toBe(1);
      expect(global.setTimeout.mock.calls.length).toBe(1);
    });

    it('resets mock setInterval after executing callback', function() {
      var nativeSetInterval = jest.genMockFn();
      var global = {setInterval: nativeSetInterval};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runWithRealTimers(function() {
        global.setInterval();
      });
      expect(nativeSetInterval.mock.calls.length).toBe(1);
      expect(global.setInterval.mock.calls.length).toBe(0);

      global.setInterval();
      expect(nativeSetInterval.mock.calls.length).toBe(1);
      expect(global.setInterval.mock.calls.length).toBe(1);
    });

    it('resets mock clearTimeout after executing callback', function() {
      var nativeClearTimeout = jest.genMockFn();
      var global = {clearTimeout: nativeClearTimeout};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runWithRealTimers(function() {
        global.clearTimeout();
      });
      expect(nativeClearTimeout.mock.calls.length).toBe(1);
      expect(global.clearTimeout.mock.calls.length).toBe(0);

      global.clearTimeout();
      expect(nativeClearTimeout.mock.calls.length).toBe(1);
      expect(global.clearTimeout.mock.calls.length).toBe(1);
    });

    it('resets mock clearInterval after executing callback', function() {
      var nativeClearInterval = jest.genMockFn();
      var global = {clearInterval: nativeClearInterval};
      var fakeTimers = new FakeTimers(global);

      fakeTimers.runWithRealTimers(function() {
        global.clearInterval();
      });
      expect(nativeClearInterval.mock.calls.length).toBe(1);
      expect(global.clearInterval.mock.calls.length).toBe(0);

      global.clearInterval();
      expect(nativeClearInterval.mock.calls.length).toBe(1);
      expect(global.clearInterval.mock.calls.length).toBe(1);
    });

    it('resets mock timer functions even if callback throws', function() {
      var nativeSetTimeout = jest.genMockFn();
      var global = {setTimeout: nativeSetTimeout};
      var fakeTimers = new FakeTimers(global);

      expect(function() {
        fakeTimers.runWithRealTimers(function() {
          global.setTimeout();
          throw new Error('test');
        });
      }).toThrow('test');
      expect(nativeSetTimeout.mock.calls.length).toBe(1);
      expect(global.setTimeout.mock.calls.length).toBe(0);

      global.setTimeout();
      expect(nativeSetTimeout.mock.calls.length).toBe(1);
      expect(global.setTimeout.mock.calls.length).toBe(1);
    });
  });
});
